#!/usr/bin/env python
# -*- coding: UTF-8 -*-
# vim: set fileencoding=UTF-8 :

__author__ = "JaeHyuk Cho <minzkn@minzkn.com>"
__date__ = "2019/11/13"
__version__ = "1.0.1.0"

import os
import sys
import platform
import inspect
#import copy
#import struct
import ctypes # ALTERNATIVE: from ctypes import *
#import socket
#import re

# 라이브러리 동적 로딩 참조 카운트 (hwport_pgl_core 내부 사용)
_g_hwport_pgl_core_dynamic_loader_reference_counter = 0

# 라이브러리 인터페이스 명
if platform.system().lower() == "windows": # Windows
    _g_hwport_pgl_default_library_pathname_tuple = (
        u"libhwport_pgl.dll",
        u"libhyper-pgl.dll",
    )
    _g_hwport_pgl_default_gbox_pathname_tuple = (
        u"libgbox.dll",
    )
elif platform.system().lower() == "darwin": # Darwin
    _g_hwport_pgl_default_library_pathname_tuple = (
        u"libhwport_pgl.dylib",
        u"libhyper-pgl.dylib",
    )
    _g_hwport_pgl_default_gbox_pathname_tuple = (
        u"libgbox.dylib",
    )
else: # Linux
    _g_hwport_pgl_default_library_pathname_tuple = (
        u"libhyper-pgl.so.0",
        u"libhwport_pgl.so.0",
        u"libhyper-pgl.so",
        u"libhwport_pgl.so",
        u"/usr/lib/libhyper-pgl.so.0",
        u"/usr/lib/libhwport_pgl.so.0",
    )
    _g_hwport_pgl_default_gbox_pathname_tuple = (
        u"libgbox.so.0",
        u"libgbox.so",
        u"/usr/lib/libgbox.so.0",
        u"/usr/lib/libgbox.so",
        u"/usr/bin/gbox",
    )
_g_hwport_pgl_default_string_encoding_type = "utf-8"

class hwport_pgl_types(object):
    def __init__(self, s_verbose = False, s_debug_level = 0):
        # => verbose and debug
        self.m_verbose = bool(s_verbose)
        try: self.m_debug_level = int(s_debug_level)
        except: self.m_debug_level = 0

        # => types
        # basic types
        self.m_ctypes_void_t = None
        self.m_ctypes_char_t = ctypes.c_char
        self.m_ctypes_uchar_t = ctypes.c_ubyte
        self.m_ctypes_short_t = ctypes.c_short
        self.m_ctypes_ushort_t = ctypes.c_ushort
        self.m_ctypes_int_t = ctypes.c_int
        self.m_ctypes_uint_t = ctypes.c_uint
        self.m_ctypes_long_t = ctypes.c_long
        self.m_ctypes_ulong_t = ctypes.c_ulong
        self.m_ctypes_longlong_t = ctypes.c_longlong
        self.m_ctypes_ulonglong_t = ctypes.c_ulonglong
        self.m_ctypes_float_t = ctypes.c_float
        self.m_ctypes_double_t = ctypes.c_double
        self.m_ctypes_longdouble_t = ctypes.c_longdouble
        self.m_ctypes_const_void_p_t = ctypes.c_void_p
        self.m_ctypes_void_p_t = ctypes.c_void_p
        self.m_ctypes_const_char_p_t = ctypes.c_char_p
        self.m_ctypes_char_p_t = ctypes.c_char_p
        self.m_ctypes_pointer_t = ctypes.pointer
        self.m_ctypes_POINTER_t = ctypes.POINTER
        self.m_ctypes_byref_t = ctypes.byref
        self.m_ctypes_LP_char_t = self.m_ctypes_POINTER_t(ctypes.c_char) # char []
        self.m_ctypes_LP_LP_char_t = self.m_ctypes_POINTER_t(self.m_ctypes_POINTER_t(ctypes.c_char))
        # 1'st representation types
        self.m_ctypes_int8_t = self.m_ctypes_char_t
        self.m_ctypes_int16_t = self.m_ctypes_short_t
        self.m_ctypes_int32_t = self.m_ctypes_int_t
        self.m_ctypes_int64_t = self.m_ctypes_longlong_t
        self.m_ctypes_intmax_t = self.m_ctypes_longlong_t
        self.m_ctypes_uint8_t = self.m_ctypes_uchar_t
        self.m_ctypes_uint16_t = self.m_ctypes_ushort_t
        self.m_ctypes_uint32_t = self.m_ctypes_uint_t
        self.m_ctypes_uint64_t = self.m_ctypes_ulonglong_t
        self.m_ctypes_uintmax_t = self.m_ctypes_ulonglong_t
        self.m_ctypes_real_t = self.m_ctypes_double_t
        self.m_ctypes_realmax_t = self.m_ctypes_longdouble_t
        self.m_ctypes_size_t = self.m_ctypes_ulong_t
        self.m_ctypes_ssize_t = self.m_ctypes_long_t
        self.m_ctypes_off_t = self.m_ctypes_longlong_t
        self.m_ctypes_fd_t = self.m_ctypes_int_t
        self.m_ctypes_socket_t = self.m_ctypes_int_t
        self.m_ctypes_atomic_t = self.m_ctypes_int_t
        # 2'st representation types
        self.m_ctypes_ptrdiff_t = self.m_ctypes_ssize_t
        """
        fourcc version code
        +-------------------------------+
        |   3   |   2   |   1   |   0   |
        |  MSB  |       |       |  LSB  |
        +~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~+
        | major | minor | extra | build |
        +-------------------------------+
        """
        self.m_hwport_fourcc_version_t = self.m_ctypes_int32_t
        return None

    def __del__(self):
        pass

    def __str__(self):
        return type(self).__name__
    def __len__(self):
        return len(self.__str__())

    def string_encode(self, s_string, s_encode_type = _g_hwport_pgl_default_string_encoding_type):
        if s_string != None:
            return s_string.encode(s_encode_type)
        return None

    def string_decode(self, s_string, s_decode_type = _g_hwport_pgl_default_string_encoding_type):
        if s_string != None:
            return s_string.decode(s_decode_type)
        return None

    def conv_argv_to_ctypes(self, s_argv = sys.argv):
        s_argc = len(s_argv)
        s_ctypes_argv = (self.m_ctypes_POINTER_t(self.m_ctypes_char_t) * (s_argc + 1))()
        for s_arg_index, s_arg in enumerate(s_argv):
            s_ctypes_argv[s_arg_index] = ctypes.create_string_buffer(self.string_encode(s_arg))
        s_ctypes_argv[s_argc] = None
        return s_ctypes_argv

class hwport_pgl_core(hwport_pgl_types):
    def __init__(self, s_verbose = False, s_debug_level = 0, s_library_pathname = None):
        global _g_hwport_pgl_core_dynamic_loader_reference_counter
        global _g_hwport_pgl_default_library_pathname_tuple
        global _g_hwport_pgl_default_gbox_pathname_tuple

        hwport_pgl_types.__init__(self, s_verbose = s_verbose, s_debug_level = s_debug_level)

        s_library_pathname_tuple = (s_library_pathname) if s_library_pathname else _g_hwport_pgl_default_library_pathname_tuple
        for s_pathname in s_library_pathname_tuple:
            try:
                self.m_pgl = ctypes.cdll.LoadLibrary(s_pathname)
                if self.m_pgl:
                    _g_hwport_pgl_core_dynamic_loader_reference_counter += 1
                    self.m_library_pathname = s_pathname
                    break
            except:
                self.m_pgl = None
        else:
            self.m_library_pathname = s_library_pathname
            hwport_pgl_types.__del__(self)
            raise Exception(u"hwport_pgl library not found !")

        for s_pathname in _g_hwport_pgl_default_gbox_pathname_tuple:
            try:
                self.m_gbox = ctypes.cdll.LoadLibrary(s_pathname)
                if self.m_gbox:
                    self.m_gbox_pathname = s_pathname
                    break
            except:
                self.m_gbox = None
        else:
            self.m_gbox_pathname = None

        # => version.c
        # const char *hwport_get_pgl_sign(void);
        self.m_pgl.hwport_get_pgl_sign.restype = self.m_ctypes_const_char_p_t
        self.m_pgl.hwport_get_pgl_sign.argtypes = []
        # const char *hwport_get_pgl_build_date(void);
        self.m_pgl.hwport_get_pgl_build_date.restype = self.m_ctypes_const_char_p_t
        self.m_pgl.hwport_get_pgl_build_date.argtypes = []
        # const char *hwport_get_pgl_build_time(void);
        self.m_pgl.hwport_get_pgl_build_time.restype = self.m_ctypes_const_char_p_t
        self.m_pgl.hwport_get_pgl_build_time.argtypes = []
        # hwport_fourcc_version_t hwport_get_pgl_version(void);
        self.m_pgl.hwport_get_pgl_version.restype = self.m_hwport_fourcc_version_t
        self.m_pgl.hwport_get_pgl_version.argtypes = []
        # const char *hwport_get_pgl_version_string(void);
        self.m_pgl.hwport_get_pgl_version_string.restype = self.m_ctypes_const_char_p_t
        self.m_pgl.hwport_get_pgl_version_string.argtypes = []
        # char *hwport_get_pgl_version_revision(void);
        self.m_pgl.hwport_get_pgl_version_revision.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_pgl_version_revision.argtypes = []
        # char *hwport_get_pgl_company(void);
        self.m_pgl.hwport_get_pgl_company.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_pgl_company.argtypes = []
        # char *hwport_get_pgl_copyrights(void);
        self.m_pgl.hwport_get_pgl_copyrights.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_pgl_copyrights.argtypes = []
        # char *hwport_get_build_username(void);
        self.m_pgl.hwport_get_build_username.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_build_username.argtypes = []
        # char *hwport_get_build_hostname(void);
        self.m_pgl.hwport_get_build_hostname.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_build_hostname.argtypes = []
        # size_t hwport_get_architecture_word_size_bits(void);
        self.m_pgl.hwport_get_architecture_word_size_bits.restype = self.m_ctypes_size_t
        self.m_pgl.hwport_get_architecture_word_size_bits.argtypes = []
        # char *hwport_get_runtime_architecture(void);
        self.m_pgl.hwport_get_runtime_architecture.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_runtime_architecture.argtypes = []
        # char *hwport_get_target_architecture(void);
        self.m_pgl.hwport_get_target_architecture.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_target_architecture.argtypes = []
        # char *hwport_get_target_build_profile(void);
        self.m_pgl.hwport_get_target_build_profile.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_target_build_profile.argtypes = []
        # char *hwport_get_compiler_name(void);
        self.m_pgl.hwport_get_compiler_name.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_compiler_name.argtypes = []
        # hwport_fourcc_version_t hwport_get_compiler_version(void);
        self.m_pgl.hwport_get_compiler_version.restype = self.m_hwport_fourcc_version_t
        self.m_pgl.hwport_get_compiler_version.argtypes = []
        # char *hwport_get_libc_name(void);
        self.m_pgl.hwport_get_libc_name.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_libc_name.argtypes = []
        # hwport_fourcc_version_t hwport_get_libc_version(void);
        self.m_pgl.hwport_get_libc_version.restype = self.m_hwport_fourcc_version_t
        self.m_pgl.hwport_get_libc_version.argtypes = []
        # char *hwport_get_os_name(void);
        self.m_pgl.hwport_get_os_name.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_os_name.argtypes = []
        # const char *hwport_get_os_version_string(void);
        self.m_pgl.hwport_get_os_version_string.restype = self.m_ctypes_const_char_p_t
        self.m_pgl.hwport_get_os_version_string.argtypes = []
        # hwport_fourcc_version_t hwport_get_os_version(void);
        self.m_pgl.hwport_get_os_version.restype = self.m_hwport_fourcc_version_t
        self.m_pgl.hwport_get_os_version.argtypes = []
        # char *hwport_get_langinfo_codeset(void);
        self.m_pgl.hwport_get_langinfo_codeset.restype = self.m_ctypes_char_p_t
        self.m_pgl.hwport_get_langinfo_codeset.argtypes = []
        # int hwport_push_version_xml(void *s_fifo, const char *s_program_name);
        self.m_pgl.hwport_push_version_xml.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_push_version_xml.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_char_p_t]

        # => base.c
        # void hwport_dump_backtrace(unsigned int s_flags, int s_priority);
        self.m_pgl.hwport_dump_backtrace.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_dump_backtrace.argtypes = [self.m_ctypes_uint_t, self.m_ctypes_int_t]
        # void hwport_set_global_break(void);
        self.m_pgl.hwport_set_global_break.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_set_global_break.argtypes = []
        # void hwport_clear_global_break(void);
        self.m_pgl.hwport_clear_global_break.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_clear_global_break.argtypes = []
        # int hwport_get_global_break(void);
        self.m_pgl.hwport_get_global_break.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_get_global_break.argtypes = []
        # void *hwport_ignore_strict_aliasing_pointer(const void *s_const_ptr);
        self.m_pgl.hwport_ignore_strict_aliasing_pointer.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_ignore_strict_aliasing_pointer.argtypes = [self.m_ctypes_const_void_p_t]

        # => nop.c
        # int hwport_nop(void);
        self.m_pgl.hwport_nop.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_nop.argtypes = []
        # int hwport_true(void);
        self.m_pgl.hwport_true.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_true.argtypes = []
        # int hwport_false(void);
        self.m_pgl.hwport_false.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_false.argtypes = []

        # => assert.c
        # void hwport_assert_fail_tag(const char *s_file, const char *s_caller_name, size_t s_caller_line, const char *s_message);
        self.m_pgl.hwport_assert_fail_tag.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_assert_fail_tag.argtypes = [self.m_ctypes_const_char_p_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t, self.m_ctypes_const_char_p_t]

        # => linked_list.c
        # void *hwport_linked_list_get_tail_tag(void **s_head_ptr, void **s_tail_ptr, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # ssize_t hwport_linked_list_index_tag(void **s_head_or_tail_ptr, void *s_element, size_t s_next_or_prev_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # ssize_t hwport_linked_list_count_tag(void **s_head_or_tail_ptr, size_t s_next_or_prev_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void *hwport_linked_list_search_tag(void **s_head_or_tail_ptr, void *s_element, size_t s_next_or_prev_offset, hwport_compare_linked_list_handler_t s_compare_handler, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL

        # => singly_linked_list.c
        # void hwport_singly_linked_list_prepend_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_singly_linked_list_append_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_singly_linked_list_delete_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_singly_linked_list_replace_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, void *s_new_element, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_singly_linked_list_insert_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, void *s_new_element, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_singly_linked_list_sort_insert_tag(void **s_head_ptr, void **s_tail_ptr, void *s_new_element, size_t s_next_offset, hwport_compare_linked_list_handler_t s_compare_handler, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_singly_linked_list_sort_tag(void **s_head_ptr, void **s_tail_ptr, size_t s_next_offset, hwport_compare_linked_list_handler_t s_compare_handler, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL

        # => doubly_linked_list.c
        # void hwport_doubly_linked_list_prepend_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, size_t s_prev_offset, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_doubly_linked_list_append_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, size_t s_prev_offset, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_doubly_linked_list_delete_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, size_t s_prev_offset, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_doubly_linked_list_replace_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, void *s_new_element, size_t s_prev_offset, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_doubly_linked_list_insert_tag(void **s_head_ptr, void **s_tail_ptr, void *s_element, void *s_new_element, size_t s_prev_offset, size_t s_next_offset, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_doubly_linked_list_sort_insert_tag(void **s_head_ptr, void **s_tail_ptr, void *s_new_element, size_t s_prev_offset, size_t s_next_offset, hwport_compare_linked_list_handler_t s_compare_handler, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL
        # void hwport_doubly_linked_list_sort_tag(void **s_head_ptr, void **s_tail_ptr, size_t s_prev_offset, size_t s_next_offset, hwport_compare_linked_list_handler_t s_compare_handler, const char *s_file, const char *s_caller_name, size_t s_caller_line);
        # NOT SUPPORTED CALL

        # => check_ptr.c
        # void *hwport_check_ptr(const void *s_const_ptr, const void *s_default_ptr);
        self.m_pgl.hwport_check_ptr.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_check_ptr.argtypes = [self.m_ctypes_const_void_p_t, self.m_ctypes_const_void_p_t]

        # => memory.c
        # void *hwport_memset8_wrapper(void *s_to, int s_value, size_t s_count);
        self.m_pgl.hwport_memset8_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memset8_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_int_t, self.m_ctypes_size_t]
        # void *hwport_memset16_wrapper(void *s_to, int s_value, size_t s_count);
        self.m_pgl.hwport_memset16_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memset16_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_int_t, self.m_ctypes_size_t]
        # void *hwport_memset32_wrapper(void *s_to, hwport_uint32_t s_value, size_t s_count);
        self.m_pgl.hwport_memset32_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memset32_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_int_t, self.m_ctypes_size_t]
        # void *hwport_memset64_wrapper(void *s_to, hwport_uint64_t s_value, size_t s_count);
        self.m_pgl.hwport_memset64_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memset64_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_int_t, self.m_ctypes_size_t]
        # void hwport_bzero_wrapper(void *s_to, size_t s_size);
        self.m_pgl.hwport_bzero_wrapper.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_bzero_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_size_t]
        # void *hwport_memset_wrapper(void *s_to, int s_value, size_t s_size);
        self.m_pgl.hwport_memset_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memset_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_int_t, self.m_ctypes_size_t]
        # void *hwport_memcpy_wrapper(void *s_to, const void *s_from, size_t s_size);
        self.m_pgl.hwport_memcpy_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memcpy_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t]
        # void *hwport_memmove_wrapper(void *s_to, const void *s_from, size_t s_size);
        self.m_pgl.hwport_memmove_wrapper.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memmove_wrapper.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t]
        # int hwport_memcmp_wrapper(const void *s_left, const void *s_right, size_t s_size);
        self.m_pgl.hwport_memcmp_wrapper.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_memcmp_wrapper.argtypes = [self.m_ctypes_const_void_p_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t]
        # void *hwport_memset_blit(void *s_to, const void *s_value_ptr, size_t s_size, size_t s_to_width, size_t s_value_size, size_t s_count);
        self.m_pgl.hwport_memset_blit.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memset_blit.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t, self.m_ctypes_size_t, self.m_ctypes_size_t, self.m_ctypes_size_t]
        # void *hwport_memcpy_blit(void *s_to, const void *s_from, size_t s_size, size_t s_to_width, size_t s_from_width, size_t s_count);
        self.m_pgl.hwport_memcpy_blit.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_memcpy_blit.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t, self.m_ctypes_size_t, self.m_ctypes_size_t, self.m_ctypes_size_t]
        # void *hwport_exchange_memory(void *s_left, void *s_right, size_t s_size);
        self.m_pgl.hwport_exchange_memory.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_exchange_memory.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_void_p_t, self.m_ctypes_size_t]
        # int hwport_is_data_fill(const void *s_data, size_t s_size, int s_value);
        self.m_pgl.hwport_is_data_fill.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_is_data_fill.argtypes = [self.m_ctypes_const_void_p_t, self.m_ctypes_size_t, self.m_ctypes_int_t]
        # int hwport_compare_data(const void *s_left_data, size_t s_left_size, const void *s_right_data, size_t s_right_size);
        self.m_pgl.hwport_compare_data.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_compare_data.argtypes = [self.m_ctypes_const_void_p_t, self.m_ctypes_size_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t]

        # => atomic_exchange.c
        # hwport_atomic_t hwport_atomic_exchange(volatile hwport_atomic_t *s_to, hwport_atomic_t s_value); /* acquire barrier */
        # void hwport_atomic_release(volatile hwport_atomic_t *s_to); /* release barrier */

        # => short_lock.c
        # hwport_short_lock_t *hwport_init_short_lock(hwport_short_lock_t *s_short_lock);
        # hwport_short_lock_t *hwport_uninit_short_lock(hwport_short_lock_t *s_short_lock);
        # int hwport_try_short_lock(hwport_short_lock_t *s_short_lock);
        # void hwport_short_lock(hwport_short_lock_t *s_short_lock);
        # void hwport_short_unlock(hwport_short_lock_t *s_short_lock);

        # => spin_lock.c
        # hwport_spin_lock_t *hwport_init_spin_lock(hwport_spin_lock_t *s_this_spin_lock);
        # hwport_spin_lock_t *hwport_uninit_spin_lock(hwport_spin_lock_t *s_this_spin_lock);
        # hwport_spin_lock_t *hwport_new_spin_lock(void);
        # hwport_spin_lock_t *hwport_free_spin_lock(hwport_spin_lock_t *s_this_spin_lock);
        # int hwport_reference_try_spin_lock_tag(hwport_spin_lock_t *s_this_spin_lock, unsigned int s_flags, const char *s_caller_name, size_t s_caller_line);
        # void hwport_reference_spin_unlock_tag(hwport_spin_lock_t *s_this_spin_lock, const char *s_caller_name, size_t s_caller_line);
        # void hwport_reference_spin_lock_tag(hwport_spin_lock_t *s_this_spin_lock, unsigned int s_flags, const char *s_caller_name, size_t s_caller_line);
        # int hwport_reference_spin_lock_timeout_tag(hwport_spin_lock_t *s_this_spin_lock, unsigned int s_flags, int s_timeout_msec, const char *s_caller_name, size_t s_caller_line);
        # int hwport_try_spin_lock_tag(hwport_spin_lock_t *s_this_spin_lock, const char *s_caller_name, size_t s_caller_line);
        # void hwport_spin_unlock_tag(hwport_spin_lock_t *s_this_spin_lock, const char *s_caller_name, size_t s_caller_line);
        # void hwport_spin_lock_tag(hwport_spin_lock_t *s_this_spin_lock, const char *s_caller_name, size_t s_caller_line);
        # int hwport_spin_lock_timeout_tag(hwport_spin_lock_t *s_this_spin_lock, int s_timeout_msec, const char *s_caller_name, size_t s_caller_line);

        # => heap.c
        # size_t __hwport_malloc_usable_size(void *s_ptr);
        # NOT SUPPORTED CALL
        # size_t hwport_get_alloc_list(size_t s_max_count, hwport_get_alloc_list_handler_t s_handler, void *s_context);
        # NOT SUPPORTED CALL
        # void *hwport_heap_tag(void *s_ptr, size_t s_size, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_heap_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_heap_tag.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_size_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void *hwport_heap_notag(void *s_ptr, size_t s_size);
        self.m_pgl.hwport_heap_notag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_heap_notag.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_size_t]
        # int hwport_get_alloc_info(size_t *s_alloc_count_ptr, size_t *s_alloc_size_ptr, size_t *s_alloc_peak_size_ptr);
        self.m_pgl.hwport_get_alloc_info.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_get_alloc_info.argtypes = [self.m_ctypes_POINTER_t(self.m_ctypes_size_t), self.m_ctypes_POINTER_t(self.m_ctypes_size_t), self.m_ctypes_POINTER_t(self.m_ctypes_size_t)]
        # size_t hwport_get_alloc_count(void);
        self.m_pgl.hwport_get_alloc_count.restype = self.m_ctypes_size_t
        self.m_pgl.hwport_get_alloc_count.argtypes = []
        # size_t hwport_get_alloc_size(void);
        self.m_pgl.hwport_get_alloc_size.restype = self.m_ctypes_size_t
        self.m_pgl.hwport_get_alloc_size.argtypes = []
        # size_t hwport_get_alloc_peak_size(void);
        self.m_pgl.hwport_get_alloc_peak_size.restype = self.m_ctypes_size_t
        self.m_pgl.hwport_get_alloc_peak_size.argtypes = []
        # void *hwport_alloc_tag(size_t s_alloc_size, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_alloc_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_alloc_tag.argtypes = [self.m_ctypes_size_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void *hwport_free_tag(void *s_ptr, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_free_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_free_tag.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void *hwport_check_free_tag(void *s_ptr, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_check_free_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_check_free_tag.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void *hwport_realloc_tag(void *s_ptr, size_t s_realloc_size, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_realloc_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_realloc_tag.argtypes = [self.m_ctypes_void_p_t, self.m_ctypes_size_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void *hwport_calloc_tag(size_t s_count, size_t s_item_size, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_calloc_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_calloc_tag.argtypes = [self.m_ctypes_size_t, self.m_ctypes_size_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]

        # => exception.c
        # void *hwport_add_exception_tag(const char *s_source_name, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_add_exception_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_add_exception_tag.argtypes = [self.m_ctypes_const_char_p_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void *hwport_del_exception_tag(void *s_handle);
        self.m_pgl.hwport_del_exception_tag.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_del_exception_tag.argtypes = [self.m_ctypes_void_p_t]
        # void hwport_dump_exception_tag(int s_force, const char *s_source_name, const char *s_caller_name, size_t s_caller_line);
        self.m_pgl.hwport_dump_exception_tag.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_dump_exception_tag.argtypes = [self.m_ctypes_int_t, self.m_ctypes_const_char_p_t, self.m_ctypes_const_char_p_t, self.m_ctypes_size_t]
        # void hwport_set_force_dump_exception(int s_force);
        self.m_pgl.hwport_set_force_dump_exception.restype = self.m_ctypes_void_t
        self.m_pgl.hwport_set_force_dump_exception.argtypes = [self.m_ctypes_int_t]

        # => puts.c
        # int hwport_quiet_puts(const char *s_string);
        self.m_pgl.hwport_quiet_puts.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_quiet_puts.argtypes = [self.m_ctypes_const_char_p_t]
        # int hwport_log_puts(const char *s_string);
        self.m_pgl.hwport_log_puts.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_log_puts.argtypes = [self.m_ctypes_const_char_p_t]
        # int hwport_error_puts(const char *s_string);
        self.m_pgl.hwport_error_puts.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_error_puts.argtypes = [self.m_ctypes_const_char_p_t]
        # int hwport_puts(const char *s_string);
        self.m_pgl.hwport_puts.restype = self.m_ctypes_int_t
        self.m_pgl.hwport_puts.argtypes = [self.m_ctypes_const_char_p_t]
        # void *hwport_dump_ex(int s_left, size_t s_seek_offset, const void *s_data, size_t s_size, hwport_off_t s_display_offset);
        self.m_pgl.hwport_dump_ex.restype = self.m_ctypes_void_p_t
        self.m_pgl.hwport_dump_ex.argtypes = [self.m_ctypes_int_t, self.m_ctypes_size_t, self.m_ctypes_const_void_p_t, self.m_ctypes_size_t, self.m_ctypes_off_t]

        if self.m_gbox:
            # => gbox_main.c
            # int hwport_multicall_gbox_top_main(int s_argc, char **s_argv);
            self.m_gbox.hwport_multicall_gbox_top_main.restype = self.m_ctypes_int_t
            self.m_gbox.hwport_multicall_gbox_top_main.argtypes = [self.m_ctypes_int_t, self.m_ctypes_LP_LP_char_t]

        if self.m_verbose or self.m_debug_level > 0:
            self.early_puts(u"hwport_pgl_core: initialized %s (verbose=%s, debug_level=%d)\n"%(self.m_library_pathname, self.m_verbose, self.m_debug_level))
            if self.m_gbox:
                self.early_puts(u"hwport_pgl_core: initialized %s\n"%(self.m_gbox_pathname))
        return None

    def __del__(self):
        global _g_hwport_pgl_core_dynamic_loader_reference_counter

        if self.m_gbox:
            if self.m_verbose or self.m_debug_level > 0:
                self.early_puts(u"hwport_pgl_core: destroy %s\n"%(self.m_gbox_pathname))
            del self.m_gbox
            self.m_gbox = None
        if self.m_pgl:
            _g_hwport_pgl_core_dynamic_loader_reference_counter -= 1
            s_hwport_pgl_core_dynamic_loader_reference_counter = _g_hwport_pgl_core_dynamic_loader_reference_counter
            if self.m_verbose or self.m_debug_level > 0:
                self.early_puts(u"hwport_pgl_core: destroy %s\n"%(self.m_library_pathname))
            if s_hwport_pgl_core_dynamic_loader_reference_counter == 0:
                # hwport heap memory leak check
                s_stack = inspect.stack()[1]
                s_frame = s_stack[0]
                s_info = inspect.getframeinfo(s_frame)
                s_is_force = 0 # 0=if leak then dump , 1=force dump
                self.m_pgl.hwport_dump_exception_tag(s_is_force, self.string_encode(s_info.filename), self.string_encode(s_info.function), s_info.lineno)
            del self.m_pgl
            self.m_pgl = None
        hwport_pgl_types.__del__(self)

    def __str__(self):
        if self.m_library_pathname:
            return self.m_library_pathname
        return type(self).__name__
    def __len__(self):
        return len(self.__str__())

    def early_puts(self, s_string):
        if s_string:
            try:
                sys.stdout.write(s_string)
            except:
                print(s_string)
            return len(self.string_encode(s_string))
        return 0

    def early_error_puts(self, s_string):
        if s_string:
            try:
                sys.stderr.write(s_string)
            except:
                print(s_string)
            return len(self.string_encode(s_string))
        return 0

    def load_library(self, s_library_pathname):
        try: s_handle = ctypes.cdll.LoadLibrary(s_library_pathname)
        except: s_handle = None
        return s_handle

    def unload_library(s_handle):
        if s_handle: del s_handle
        return None

class hwport_pgl(hwport_pgl_core):
    def __init__(self, s_verbose = False, s_debug_level = 0, s_library_pathname = None):
        hwport_pgl_core.__init__(self, s_verbose = s_verbose, s_debug_level = s_debug_level, s_library_pathname = s_library_pathname)
        return None

    def __del__(self):
        hwport_pgl_core.__del__(self)

    # => version.c
    def get_pgl_sign(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_sign())
    def get_pgl_build_date(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_build_date())
    def get_pgl_build_time(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_build_time())
    def get_pgl_version(self):
        return self.m_pgl.hwport_get_pgl_version()
    def get_pgl_version_string(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_version_string())
    def get_pgl_version_revision(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_version_revision())
    def get_pgl_company(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_company())
    def get_pgl_copyrights(self):
        return self.string_decode(self.m_pgl.hwport_get_pgl_copyrights())
    def get_build_username(self):
        return self.string_decode(self.m_pgl.hwport_get_build_username())
    def get_build_hostname(self):
        return self.string_decode(self.m_pgl.hwport_get_build_hostname())
    def get_architecture_word_size_bits(self):
        return self.m_pgl.hwport_get_architecture_word_size_bits();
    def get_runtime_architecture(self):
        return self.string_decode(self.m_pgl.hwport_get_runtime_architecture())
    def get_target_architecture(self):
        return self.string_decode(self.m_pgl.hwport_get_target_architecture())
    def get_target_build_profile(self):
        return self.string_decode(self.m_pgl.hwport_get_target_build_profile())
    def get_compiler_name(self):
        return self.string_decode(self.m_pgl.hwport_get_compiler_name())
    def get_compiler_version(self):
        return self.m_pgl.hwport_get_compiler_version()
    def get_compiler_version_string(self):
        s_fourcc = self.m_pgl.hwport_get_compiler_version()
        if s_fourcc & 0xff:
            return "%u.%u.%u-%u"%((s_fourcc >> 24) & 0xff, (s_fourcc >> 16) & 0xff, (s_fourcc >> 8) & 0xff, s_fourcc & 0xff)
        else:
            return "%u.%u.%u"%((s_fourcc >> 24) & 0xff, (s_fourcc >> 16) & 0xff, (s_fourcc >> 8) & 0xff)
    def get_libc_name(self):
        return self.string_decode(self.m_pgl.hwport_get_libc_name())
    def get_libc_version(self):
        return self.m_pgl.hwport_get_libc_version()
    def get_libc_version_string(self):
        s_fourcc = self.m_pgl.hwport_get_libc_version()
        if s_fourcc & 0xff:
            return "%u.%u.%u-%u"%((s_fourcc >> 24) & 0xff, (s_fourcc >> 16) & 0xff, (s_fourcc >> 8) & 0xff, s_fourcc & 0xff)
        else:
            return "%u.%u.%u"%((s_fourcc >> 24) & 0xff, (s_fourcc >> 16) & 0xff, (s_fourcc >> 8) & 0xff)
    def get_os_name(self):
        return self.string_decode(self.m_pgl.hwport_get_os_name())
    def get_os_version_string(self):
        return self.string_decode(self.m_pgl.hwport_get_os_version_string())
    def get_os_version(self):
        return self.m_pgl.hwport_get_os_version()
    def get_langinfo_codeset(self):
        return self.string_decode(self.m_pgl.hwport_get_langinfo_codeset())
    def push_version_xml(self, s_fifo, s_program_name):
        return self.m_pgl.hwport_push_version_xml(s_fifo, self.string_encode(s_program_name))

    # => base.c
    def dump_backtrace(self, s_flags = 0x0000002, s_priority = 0):
        self.m_pgl.hwport_dump_backtrace(s_flags, s_priority)
        return None
    def set_global_break(self):
        self.m_pgl.hwport_set_global_break()
        return None
    def clear_global_break(self):
        self.m_pgl.hwport_clear_global_break()
        return None
    def get_global_break(self):
        return self.m_pgl.hwport_get_global_break()
    def ignore_strict_aliasing_pointer(self, s_const_ptr):
        return self.m_pgl.hwport_ignore_strict_aliasing_pointer(s_const_ptr)

    # => nop.c
    def nop(self):
        return self.m_pgl.hwport_nop()
    def true(self):
        return self.m_pgl.hwport_true()
    def false(self):
        return self.m_pgl.hwport_false()

    # => assert.c
    def assert_fail(self, s_message = u"FAIL", s_file = None, s_caller_name = None, s_caller_line = None):
        if bool(s_message) == False:
            s_message = u""
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_assert_fail_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line, self.string_encode(s_message))
        return None
    def bugon(self, s_message = u"BUG", s_file = None, s_caller_name = None, s_caller_line = None):
        if bool(s_message) == False:
            s_message = u""
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_assert_fail_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line, self.string_encode(s_message))
        return None
    def assert_puts(self, s_bool, s_message = u"FAIL", s_file = None, s_caller_name = None, s_caller_line = None):
        if bool(s_bool):
            return None
        if bool(s_message) == False:
            s_message = u""
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_assert_fail_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line, self.string_encode(s_message))
        return None
    def assert_check(self, s_bool, s_file = None, s_caller_name = None, s_caller_line = None):
        if bool(s_bool):
            return None
        s_message = str(s_bool)
        if bool(s_message) == False:
            s_message = u""
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_assert_fail_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line, self.string_encode(s_message))
        return None
    def todo(self, s_file = None, s_caller_name = None, s_caller_line = None):
        s_message = u"TODO"
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_assert_fail_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line, self.string_encode(s_message))
        return None
    def not_supported(self, s_file = None, s_caller_name = None, s_caller_line = None):
        s_message = u"not supported"
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_assert_fail_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line, self.string_encode(s_message))
        return None

    # => linked_list.c

    # => singly_linked_list.c

    # => doubly_linked_list.c

    # => check_ptr.c
    def check_ptr(self, s_ptr, s_default_ptr = None):
        return self.m_pgl.hwport_check_ptr(s_ptr, s_default_ptr)

    # => memory.c
    def memset8(self, s_to, s_value, s_count):
        return self.m_pgl.hwport_memset8_wrapper(s_to, s_value, s_count)
    def memset16(self, s_to, s_value, s_count):
        return self.m_pgl.hwport_memset16_wrapper(s_to, s_value, s_count)
    def memset32(self, s_to, s_value, s_count):
        return self.m_pgl.hwport_memset32_wrapper(s_to, s_value, s_count)
    def memset64(self, s_to, s_value, s_count):
        return self.m_pgl.hwport_memset64_wrapper(s_to, s_value, s_count)
    def bzero(self, s_to, s_size):
        self.m_pgl.hwport_bzero_wrapper(s_to, s_size)
        return s_to
    def memset(self, s_to, s_value, s_size):
        return self.m_pgl.hwport_memset_wrapper(s_to, s_value, s_size)
    def memcpy(self, s_to, s_from, s_size):
        return self.m_pgl.hwport_memcpy_wrapper(s_to, s_from, s_size)
    def memmove(self, s_to, s_from, s_size):
        return self.m_pgl.hwport_memmove_wrapper(s_to, s_from, s_size)
    def memcmp(self, s_left, s_right, s_size):
        return self.m_pgl.hwport_memcmp_wrapper(s_left, s_right, s_size)
    def memset_blit(self, s_to, s_value_ptr, s_size, s_to_width, s_value_size, s_count):
        return self.m_pgl.hwport_memset_blit(s_to, s_value_ptr, s_size, s_to_width, s_value_size, s_count)
    def memcpy_blit(self, s_to, s_from, s_size, s_to_width, s_from_width, s_count):
        return self.m_pgl.hwport_memcpy_blit(s_to, s_from, s_size, s_to_width, s_from_width, s_count)
    def exchange_memory(self, s_left, s_right, s_size):
        return self.m_pgl.hwport_exchange_memory(s_left, s_right, s_size)
    def is_data_fill(self, s_data, s_size, s_value):
        return self.m_pgl.hwport_is_data_fill(s_data, s_size, s_value)
    def compare_data(self, s_left_data, s_left_size, s_right_data, s_right_size):
        return self.m_pgl.hwport_compare_data(s_left_data, s_left_size, s_right_data, s_right_size)

    # => atomic_exchange.c

    # => short_lock.c

    # => spin_lock.c

    # => heap.c
    def heap(self, s_ptr, s_size, s_caller_name = None, s_caller_line = None):
        if s_size == None:
            s_size = 0
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_heap_tag(s_ptr, s_size, self.string_encode(s_caller_name), s_caller_line)
    def heap_notag(self, s_ptr, s_size):
        if s_size == None:
            s_size = 0
        return self.m_pgl.hwport_heap_notag(s_ptr, s_size)
    def get_alloc_info(self):
        s_alloc_count = self.m_ctypes_size_t()
        s_alloc_size = self.m_ctypes_size_t()
        s_alloc_peak_size = self.m_ctypes_size_t()
        if self.m_pgl.hwport_get_alloc_info(self.m_ctypes_byref_t(s_alloc_count), self.m_ctypes_byref_t(s_alloc_size), self.m_ctypes_byref_t(s_alloc_peak_size)) == 0:
            return (s_alloc_count.value, s_alloc_size.value, s_alloc_peak_size.value)
        return None
    def get_alloc_count(self):
        return self.m_pgl.hwport_get_alloc_count()
    def get_alloc_size(self):
        return self.m_pgl.hwport_get_alloc_size()
    def get_alloc_peak_size(self):
        return self.m_pgl.hwport_get_alloc_peak_size()
    def alloc(self, s_size, s_caller_name = None, s_caller_line = None):
        if s_size == None:
            s_size = 0
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_alloc_tag(s_size, self.string_encode(s_caller_name), s_caller_line)
    def free(self, s_ptr, s_caller_name = None, s_caller_line = None):
        if s_ptr == None:
            return None
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_free_tag(s_ptr, self.string_encode(s_caller_name), s_caller_line)
    def check_free(self, s_ptr, s_caller_name = None, s_caller_line = None):
        if s_ptr == None:
            return None
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_check_free_tag(s_ptr, self.string_encode(s_caller_name), s_caller_line)
    def realloc(self, s_ptr, s_realloc_size, s_caller_name = None, s_caller_line = None):
        if s_realloc_size == None:
            s_realloc_size = 0
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_realloc_tag(s_ptr, s_realloc_size, self.string_encode(s_caller_name), s_caller_line)
    def calloc(self, s_count, s_item_size, s_caller_name = None, s_caller_line = None):
        if s_count == None:
            s_count = 0
        if s_item_size == None:
            s_item_size = 0
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_calloc_tag(s_count, s_item_size, self.string_encode(s_caller_name), s_caller_line)

    # => exception.c
    def add_exception(self, s_file = None, s_caller_name = None, s_caller_line = None):
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        return self.m_pgl.hwport_add_exception_tag(self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line)
    def del_exception(self, s_handle):
        if s_handle == None:
            return None
        return self.m_pgl.hwport_del_exception_tag(s_handle)
    def dump_exception(self, s_force = True, s_file = None, s_caller_name = None, s_caller_line = None):
        s_stack = inspect.stack()[1]
        s_frame = s_stack[0]
        s_info = inspect.getframeinfo(s_frame)
        if bool(s_file) == False:
            s_file = s_info.filename
        if bool(s_caller_name) == False:
            s_caller_name = s_info.function
        if bool(s_caller_line) == False:
            s_caller_line = s_info.lineno
        self.m_pgl.hwport_dump_exception_tag(1 if bool(s_force) else 0, self.string_encode(s_file), self.string_encode(s_caller_name), s_caller_line)
        return None
    def set_force_dump_exception(self, s_force = True):
        self.m_pgl.hwport_set_force_dump_exception(1 if bool(s_force) else 0)
        return None

    # => puts.c
    def quiet_puts(self, s_string):
        if bool(self.m_pgl):
            return self.m_pgl.hwport_quiet_puts(self.string_encode(s_string))
        return len(self.string_encode(s_string))
    def log_puts(self, s_string):
        if bool(self.m_pgl):
            return self.m_pgl.hwport_log_puts(self.string_encode(s_string))
        return 0
    def error_puts(self, s_string):
        if bool(self.m_pgl):
            return self.m_pgl.hwport_error_puts(self.string_encode(s_string))
        return self.early_error_puts(s_string)
    def puts(self, s_string):
        if bool(self.m_pgl):
            return self.m_pgl.hwport_puts(self.string_encode(s_string))
        return self.early_puts(s_string)
    def hexa_dump(self, s_data, s_title = None, s_left = None, s_seek_offset = None, s_size = None, s_display_offset = None):
        if s_data == None: # unlikely condition
            return None
        if s_title != None:
            self.puts(s_title)
        try:
            s_data_size = len(s_data)
        except:
            s_data_size = None
        if (s_left == None) or (s_left < 0): # default
            s_left = 0
        if (s_seek_offset == None) or (s_seek_offset < 0): # default
            s_seek_offset = 0
        if (s_size == None) or (s_size < 0): # default
            s_size = s_data_size if s_data_size != none else 0
        if s_data_size == None:
            s_data_size = s_size
        if (s_display_offset == None) or (s_display_offset < 0): # default
            s_display_offset = s_seek_offset
        if s_seek_offset >= s_data_size: # unlikely condition
            return s_data
        if (s_seek_offset + s_size) > s_data_size: # unlikely condition
            s_size = s_data_size - s_seek_offset
        return self.m_pgl.hwport_dump_ex(s_left, s_seek_offset, s_data, s_size, s_display_offset)

    # => gbox_main.c
    def gbox_top_main(self, s_argc = len(sys.argv), s_argv = sys.argv, s_env = os.environ):
        if self.m_gbox:
            return self.m_gbox.hwport_multicall_gbox_top_main(s_argc, self.conv_argv_to_ctypes(s_argv))
        return 1

    def unit_test(self):
        try:
            self.early_puts(u"BEGIN: SELF TEST\n")

            # => puts.c test
            self.quiet_puts(u"quiet_puts test: %s\n"%(u"OK"))
            #self.log_puts("log_puts test: OK\n")
            self.error_puts(u"error_puts test: %s\n"%(u"OK"))
            s_test_string = u"한글"
            self.puts(u"puts test: %s (%s %s, %s: len vs elen = %d vs %d)\n"%(u"OK", u"한글", u"테스트", s_test_string, len(s_test_string), len(self.string_encode(s_test_string))))

            # => version.c test
            self.puts(u"name: \"%s\" (len=%d)\n"%(self, len(self)))
            #self.puts(self.get_pgl_sign())
            self.puts(u"\t- build time: \"%s %s\"\n"%(self.get_pgl_build_date(), self.get_pgl_build_time()))
            self.puts(u"\t- version: 0x%08X (\"%s\")\n"%(self.get_pgl_version(), self.get_pgl_version_string()))
            self.puts(u"\t- revision: \"%s\"\n"%(self.get_pgl_version_revision()))
            self.puts(u"\t- company: \"%s\"\n"%(self.get_pgl_company()))
            self.puts(u"\t- copyrights: \"%s\"\n"%(self.get_pgl_copyrights()))
            self.puts(u"\t- build username: \"%s\"\n"%(self.get_build_username()))
            self.puts(u"\t- build hostname: \"%s\"\n"%(self.get_build_hostname()))
            self.puts(u"\t- architecture word size: %lu bits\n"%(self.get_architecture_word_size_bits()))
            self.puts(u"\t- runtime architecture: \"%s\"\n"%(self.get_runtime_architecture()))
            self.puts(u"\t- target architecture: \"%s\"\n"%(self.get_target_architecture()))
            self.puts(u"\t- target build profile: \"%s\"\n"%(self.get_target_build_profile()))
            self.puts(u"\t- compiler: \"%s\" (version=0x%08X [\"%s\"])\n"%(self.get_compiler_name(), self.get_compiler_version(), self.get_compiler_version_string()))
            self.puts(u"\t- libc: \"%s\" (version=0x%08X [\"%s\"])\n"%(self.get_libc_name(), self.get_libc_version(), self.get_libc_version_string()))
            self.puts(u"\t- os: \"%s\" (version=0x%08X [\"%s\"])\n"%(self.get_os_name(), self.get_os_version(), self.get_os_version_string()))
            self.puts(u"\t- langinfo codeset: \"%s\"\n"%(self.get_langinfo_codeset()))

            self.hexa_dump(self.string_encode(u"hexa dump test 한글 END"), s_title=u"HEXA DUMP TITLE\n===============\n", s_left = 4, s_seek_offset = 3, s_size = 1000, s_display_offset = 0x1234)

            # => base.c test
            self.dump_backtrace(2, 0)
            self.puts(u"global break: %d\n"%(self.get_global_break()))
            self.set_global_break()
            self.puts(u"set global break: %d\n"%(self.get_global_break()))
            self.clear_global_break()
            self.puts(u"clear global break: %d\n"%(self.get_global_break()))
            self.puts(u"ignore_strict_aliasing_pointer: 0x%08lX\n"%(self.ignore_strict_aliasing_pointer(self.string_encode(u""))))

            # => nop.c test
            self.puts(u"nop: %d\n"%(self.nop()))
            self.puts(u"true: %d\n"%(self.true()))
            self.puts(u"false: %d\n"%(self.false()))

            # => assert.c test
            """
            self.assert_fail(u"assert_fail TEST: %s"%(u"OK"))
            self.bugon()
            self.assert_puts(True, u"assert_puts(False) TEST: %s"%(u"FAIL"))
            self.assert_puts(False, u"assert_puts(False) TEST: %s"%(u"OK"))
            self.assert_check(True, u"assert_check(True) TEST: %s"%(u"FAIL"))
            self.assert_check(False, u"assert_check(False) TEST: %s"%(u"OK"))
            self.todo()
            self.not_supported()
            """

            # => linked_list.c test

            # => singly_linked_list.c test

            # => doubly_linked_list.c test

            # => check_ptr.c test
            # => memory.c test
            # => atomic_exchange.c test
            # => short_lock.c test
            # => spin_lock.c test
            # => heap.c test
            # => exception.c test
            self.set_force_dump_exception(True)
            s_exception_handle = self.add_exception()
            s_ptr1 = self.heap(None, 100)
            s_ptr2 = self.heap_notag(None, 100)
            s_ptr3 = self.alloc(100)
            s_ptr4 = self.realloc(None, 200)
            s_ptr5 = self.calloc(2, 128)
            s_ptr6 = self.check_ptr(s_ptr5)
            s_ptr7 = self.check_ptr(None, s_ptr6)
            self.puts("ptr: 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx 0x%08lx\n"%(s_ptr1, s_ptr2, s_ptr3, s_ptr4, s_ptr5, s_ptr6, s_ptr7))
            self.hexa_dump(
                self.memset8(s_ptr1, 0x5a, 100),
                s_title = u"memset8 by 0x5a\n",
                s_left = 4,
                s_seek_offset = 0,
                s_size = 100,
                s_display_offset = 0
            )
            self.hexa_dump(
                self.memset16(s_ptr1, 0x5aa5, 100 >> 1),
                s_title = u"memset16 by 0x5aa5\n",
                s_left = 4,
                s_seek_offset = 0,
                s_size = 100,
                s_display_offset = 0
            )
            self.hexa_dump(
                self.memset32(s_ptr1, 0x5aa51234, 100 >> 2),
                s_title = u"memset32 by 0x5aa51234\n",
                s_left = 4,
                s_seek_offset = 0,
                s_size = 100,
                s_display_offset = 0
            )
            self.hexa_dump(
                self.memset64(s_ptr1, 0x5aa5123487654321, 100 >> 3),
                s_title = u"memset64 by 0x5aa5123487654321\n",
                s_left = 4,
                s_seek_offset = 0,
                s_size = 100,
                s_display_offset = 0
            )
            self.hexa_dump(
                self.bzero(s_ptr1, 100),
                s_title = u"bzero\n",
                s_left = 4,
                s_seek_offset = 0,
                s_size = 100,
                s_display_offset = 0
            )
            self.hexa_dump(
                self.memmove(
                    s_ptr3,
                    self.memcpy(
                        s_ptr2,
                        self.memset(s_ptr1, 0x82, 100),
                        100
                    ),
                    100
                ),
                s_title = u"memcpy\n",
                s_left = 4,
                s_seek_offset = 0,
                s_size = 100,
                s_display_offset = 0
            )
            self.puts("memcmp ptr1 == ptr2 ? %d\n"%(self.memcmp(s_ptr1, s_ptr2, 100)))
            # TODO TEST : self.memset_blit
            # TODO TEST : self.memcpy_blit
            # TODO TEST : self.exchange_memory
            # TODO TEST : self.is_data_fill
            # TODO TEST : self.compare_data
            self.dump_exception(True)
            self.puts("alloc count: %lu\n"%(self.get_alloc_count()))
            self.puts("alloc size: %lu\n"%(self.get_alloc_size()))
            self.puts("alloc peak size: %lu\n"%(self.get_alloc_peak_size()))
            self.puts("get alloc info: count=%lu, size=%lu, peak_size=%lu\n"%(self.get_alloc_info()))
            s_ptr5 = self.check_free(s_ptr5)
            s_ptr4 = self.check_free(s_ptr4)
            s_ptr3 = self.free(s_ptr3)
            s_ptr2 = self.free(s_ptr2)
            s_ptr1 = self.free(s_ptr1)
            s_exception_handle = self.del_exception(s_exception_handle)

            self.puts(None) # flush
            self.early_puts(u"END: SELF TEST\n")
        except Exception as s_message:
            self.early_error_puts(u"ERROR: %s\n"%(s_message))
            return False

        return True

    # main.c
    def main(self, s_argc = len(sys.argv), s_argv = sys.argv, s_env = os.environ):
        s_exit_code = None

        if s_argc >= 2:
            if s_argv[1] == u"unit_test":
                self.m_verbose = True
                self.m_debug_level = 1
                if self.unit_test():
                    s_exit_code = 0
                    self.early_puts(u"UNIT TEST SUCCESS.\n")
                else:
                    s_exit_code = 1
                    self.early_error_puts(u"UNIT TEST FAILED !\n")
    
        if s_exit_code == None:
            # call gbox main
            s_exit_code = self.gbox_top_main(s_argc = s_argc, s_argv = s_argv, s_env = s_env)

        return s_exit_code

# === ENTRY ===

if __name__ == "__main__":
    sys.exit(hwport_pgl().main(s_argc = len(sys.argv), s_argv = sys.argv, s_env = os.environ))

# End of hwport_pgl.py
